#!/usr/bin/env ruby

require 'net/https'
require 'uri'

module Gmail
    URL_GMAIL = "http://gmail.google.com/gmail"

    MBOX_SENT =		'sent'
    MBOX_ALL =		'all'
    MBOX_INBOX =	'inbox'
    MBOX_SPAM =		'spam'
    MBOX_STARRED =	'starred'

    module Utils
	def Utils.urlencode(kv)
	    str = ""
	    kv.each { |k, v|
		str += "&#{k}=#{v}"
	    }
	    URI.encode(str)
	end

	def Utils.parse_ruby_ary(str_ary, idx = 0)
	    new_ary = []
	    elem = ''

	    str_ary.strip
	    idx += 1 if str_ary[0].chr == '[' 
	    
	    while (idx < str_ary.size) do
		b = str_ary[idx].chr
	
		case (b) 
		when '['
		    idx, elem = parse_ruby_ary(str_ary, idx)
		    
		when ']'
		    break

		when '"'
		    while (idx + 1 < str_ary.size) do
			idx += 1
			b = str_ary[idx].chr
			
			if b == '"' and str_ary[idx - 1].chr != '\\' then
			    break
			end
			elem << b
		    end
		    
		when ','
		    new_ary << elem
		    elem = ''

		else
		    elem << b
		    
		end
		idx += 1
	    end
	    new_ary << elem
	    [ idx, new_ary ]
	end
	
    end

    class LiveCookie
	attr_reader :cookies
	
	def initialize
	    @cookies = Hash.new
	end
	
	def add_raw(raw_cookie)
	    raw_cookie =~ /(\w+)=([^|;]*)/
	    @cookies[$1] = $2
	end

	def dump
	    str = ""
	    @cookies.each { |k, v|
		str += "#{k}=#{v}; "
	    }
	    str.chomp("; ")
	end

	alias << add_raw
    end
    
    class Account
	class GmailLoginFailure < StandardError; end

	def initialize(username, password, http_proxy = nil)
	    @live_cookie = LiveCookie.new

	    @username = username
	    @password = password
	end

	def show_headers(resp)
	    resp.each { |k, v|
		puts "#{k}: #{v}"
	    }
	end

	def check_location(resp)
	    resp.each { |k, v|
		if (k == 'location') then
		    return v
		end
	    }	    
	end

	def store_cookies(resp)
	    resp.each { |k, v|
		if k == 'set-cookie' then
		    v.split(/, (?=\w+=)/).each { |cookie|
			@live_cookie << cookie
		    }
		end
	    }
	end

	def login
	    http = Net::HTTP.new("www.google.com", 443)
	    http.use_ssl = true
	    http.verify_mode = OpenSSL::SSL::VERIFY_NONE

	    sd = {
		'continue' => URL_GMAIL,
		'service' => 'mail',
		'Email' => @username,
		'Passwd' => @password
	    }

	    hdrs = {
		'User-Agent' => 'RubyGmailApp; Mozilla/5.0 (compatible;)'
	    }

	    resp, rd = http.post("/accounts/ServiceLoginBoxAuth", 
				 Utils.urlencode(sd), hdrs);
	    rd =~ /top\.location\W=\W"CheckCookie\?continue=([^|"]+)/
	    begin
		redir_url = URI.decode($1)
	    rescue
		raise GmailLoginFailure, "Login Failed"
	    end
	    store_cookies(resp)
	    resp, rd = retrive(redir_url)
	end

	def parse_d(str)
	    ds = [] 
	    
	    str.scan(/D\((.*?)\);/) { |d|
		ds << $1
	    }
	    ds
	end

	def parse_qu(ary)
	    @quota = "#{ary[1]} of #{ary[2]}"
	end

	load 'ymhtml.rb'
	
	def parse_thread_spans(str)
	    stream = YmHTML::InputStream.new("<body>#{str}</body>")
	    p = YmHTML::Parser.new
	    
	    addrs = []
	    p.parse(stream) do |t, n, d|
		if t == :START_ELEM and n == 'span'
		    addrs << d['id'][6..d['id'].size]
		end
	    end
	    addrs
	end

	def replace_html_content(html)
	    stream = YmHTML::InputStream.new("<body>#{html}</body>")
	    p = YmHTML::Parser.new	    

	    p.parse(stream) do |t, n, d|
		if t == :START_ELEM and n == 'span'
		    addrs << d['id'][6..d['id'].size]
		end
	    end
	end

	def parse_threads(ary)
	    threads = []

	    ary.each { |e|
		next if e == 't'

		addrs = parse_thread_spans(e[4])
		threads = [e[0], addrs, e[6], e[7]]
	    }
	    threads
	end

	def query_contacts
	    url = URL_GMAIL + '?view=cl&search=contacts'
#	    url = URL_GMAIL + '?&ik=7b5adde052&view=cl&search=contacts&pnl=d'
	    
	    resp, rd = retrive(url)
	    rd.tr!("\n", " ")

	    contacts = []

	    res = eval(parse_d(rd)[4])
	    res[1..res.size].each { |c|
		contacts << [c[1], c[3]]
	    }
	    contacts
	end

	def query_mbox(name)
	    url = URL_GMAIL + "?search=#{name}&view=tl&start=0"

	    resp, rd = retrive(url)    
	    rd.tr!("\n", " ")

	    threads = []

	    parse_d(rd).each { |d|
		m = eval(d)
		
		parse_qu(m) if m[0] == 'qu'
		threads << parse_threads(m) if m[0] == 't'
	    }	    
	    threads
	end

	def parse_mi(mi)
	    [mi[3], [mi[6], mi[7]], mi[10], mi[14], mi[15]] 
	end

	def parse_mb(mb)
	    mb[1] # Iconv.conv('koi8-r', 'utf-8', mb[1])
	end

	def open_thread(mailid)
	    url = URL_GMAIL + "?view=cv&search=inbox&th=#{mailid}"

	    thread = []
	    mbody = nil

	    resp, rd = retrive(url)
	    rd.tr!("\n", " ")

	    parse_d(rd).each { |d|
		idx, m = Utils.parse_ruby_ary(d)

		parse_qu(m) if m[0] == 'qu'
		thread << parse_mi(m) if m[0] == 'mi'
		mbody = parse_mb(m) if m[0] == 'mb'
	    }	 
	    [ thread, mbody ]
	end

	def read_mail(mailid)
	    url = URL_GMAIL + "?view=cv&search=inbox&th=#{mailid}&msgs=#{mailid}"
	    resp, rd = retrive(url)
	    rd.tr!("\n", " ")
	    
	    puts rd

	    parse_d(rd).each { |d|
		idx, m = Utils.parse_ruby_ary(d)

		parse_qu(m) if m[0] == 'qu'
		parse_mi(m) if m[0] == 'mi'
		parse_mb(m) if m[0] == 'mb'
	    }	    
	end

	def retrive(url)
	    uri = URI.parse(url)
	    
	    http = Net::HTTP.new(uri.host)
	    resp, rd = http.get(uri.path + "?" + uri.query, 
				{ "Cookie" => @live_cookie.dump })
	    store_cookies(resp)

	    [ resp, rd ]
	end
    end
end

require 'iconv'

print "Enter username: "
username = $stdin.readline.chop
if username == ''
    username = 'sunwinder@gmail.com'
end
print "Password: "
password = $stdin.readline.chop

ac = Gmail::Account.new(username, password)
ac.login

ac.query_contacts.each { |c|
    p c
}

threads = ac.query_mbox(Gmail::MBOX_INBOX)
threads.each { |thread|
    puts "TID: #{thread[0]}"
    puts " -- Subject: #{Iconv.conv('koi8-r', 'utf-8', thread[2])}"

    mails, mbody = ac.open_thread(thread[0])
    mails.each { |m|
	puts " -- #{Iconv.conv('koi8-r', 'utf-8', m[1][0])}, #{m[1][1]}"
#	puts " --- Subject: #{Iconv.conv('koi8-r', 'utf-8', m[4])}"
    }
}

# mails, mbody = ac.open_thread("101c103578b569bd")
# mails.each { |m|
#      puts "#{Iconv.conv('koi8-r', 'utf-8', m[1][0])}, #{m[1][1]}"
# }

